import Random
import Zygote
import RestrictedBoltzmannMachines as RBMs

using Test: @test, @testset, @inferred
using Statistics: mean, var, cov
using Random: bitrand
using LogExpFunctions: logistic
using QuadGK: quadgk
using RestrictedBoltzmannMachines: flatten, batch_size, batchmean, batchvar, batchcov
using RestrictedBoltzmannMachines: Binary, Spin, Potts, Gaussian, ReLU, dReLU, xReLU, pReLU
using RestrictedBoltzmannMachines: mean_from_inputs, var_from_inputs, meanvar_from_inputs,
    std_from_inputs, mean_abs_from_inputs, sample_from_inputs, mode_from_inputs
using RestrictedBoltzmannMachines: energy, free_energy, free_energies, energies

Random.seed!(2)

_layers = (
    Binary,
    Spin,
    Potts,
    Gaussian,
    ReLU,
    dReLU,
    pReLU,
    xReLU
)

random_layer(::Type{T}, N::Int...) where {T <: Union{Binary,Spin,Potts}} = T(randn(N...))
random_layer(::Type{T}, N::Int...) where {T <: Union{Gaussian,ReLU}} = T(randn(N...), randn(N...))
random_layer(::Type{dReLU}, N::Int...) = dReLU(randn(N...), randn(N...), randn(N...), randn(N...))
random_layer(::Type{xReLU}, N::Int...) = xReLU(randn(N...), randn(N...), randn(N...), randn(N...))
random_layer(::Type{pReLU}, N::Int...) = pReLU(randn(N...), randn(N...), randn(N...), 2rand(N...) .- 1)

@testset "testing $Layer" for Layer in _layers
    N = (3,2)
    layer = random_layer(Layer, N...)

    @test (@inferred size(layer)) == N
    for d in 1:length(N)
        @test (@inferred size(layer, d)) == N[d]
    end
    @test (@inferred length(layer)) == prod(N)
    @test (@inferred ndims(layer)) == length(N)
    @test size(@inferred repeat(layer,2,3,4)) == (((2, 3) .* N)..., 4)
    @test (@inferred batch_size(layer, rand(N...))) == ()
    @test (@inferred energy(layer, rand(N...))) isa Number
    @test (@inferred free_energy(layer)) isa Number
    @test (@inferred free_energy(layer, rand(N...))) isa Number
    @test size(@inferred mean_from_inputs(layer)) == size(layer)
    @test size(@inferred var_from_inputs(layer)) == size(layer)
    @test size(@inferred sample_from_inputs(layer)) == size(layer)
    @test free_energies(layer, 0) ≈ free_energies(layer)
    @test std_from_inputs(layer) ≈ sqrt.(var_from_inputs(layer))

    if layer isa RBMs.Potts
        @test size(@inferred free_energies(layer)) == (1, size(layer)[2:end]...)
    else
        @test size(@inferred free_energies(layer)) == size(layer)
    end

    @test size(@inferred sample_from_inputs(layer, 0)) == size(layer)

    for B in ((), (2,), (1,2))
        x = rand(N..., B...)
        @test (@inferred batch_size(layer, x)) == (B...,)
        @test (@inferred RBMs.batchdims(layer, x)) == (length(N) + 1):ndims(x)
        @test size(@inferred energy(layer, x)) == (B...,)
        @test size(@inferred free_energy(layer, x)) == (B...,)
        @test size(@inferred energies(layer, x)) == size(x)
        @test size(@inferred sample_from_inputs(layer, x)) == size(x)
        @test size(@inferred mean_from_inputs(layer, x)) == size(x)
        @test size(@inferred var_from_inputs(layer, x)) == size(x)
        @test @inferred(std_from_inputs(layer, x)) ≈ sqrt.(var_from_inputs(layer, x))
        @test all(energy(layer, mode_from_inputs(layer)) .≤ energy(layer, x))

        μ, ν = meanvar_from_inputs(layer, x)
        @test μ ≈ mean_from_inputs(layer, x)
        @test ν ≈ var_from_inputs(layer, x)

        if layer isa RBMs.Potts
            @test size(@inferred free_energies(layer, x)) == (1, size(x)[2:end]...)
        else
            @test size(@inferred free_energies(layer, x)) == size(x)
        end

        if B == ()
            @test @inferred(energy(layer, x)) ≈ sum(energies(layer, x))
            @test @inferred(free_energy(layer, x)) ≈ sum(free_energies(layer, x))
            @test @inferred(batchmean(layer, x)) ≈ x
            @test @inferred(batchvar(layer, x)) == zeros(N)
            @test @inferred(batchcov(layer, x)) == zeros(N..., N...)
        else
            @test @inferred(energy(layer, x)) ≈ reshape(sum(energies(layer, x); dims=1:ndims(layer)), B)
            @test @inferred(free_energy(layer, x)) ≈ reshape(sum(free_energies(layer, x); dims=1:ndims(layer)), B)
            @test @inferred(batchmean(layer, x)) ≈ reshape(mean(x; dims=(ndims(layer) + 1):ndims(x)), N)
            @test @inferred(batchvar(layer, x)) ≈ reshape(var(x; dims=(ndims(layer) + 1):ndims(x), corrected=false), N)
            @test @inferred(batchcov(layer, x)) ≈ reshape(cov(flatten(layer, x); dims=2, corrected=false), N..., N...)
        end

        μ = @inferred mean_from_inputs(layer, x)
        @test only(Zygote.gradient(j -> sum(free_energies(layer, j)), x)) ≈ -μ
    end

    ∂F = @inferred RBMs.∂free_energy(layer)
    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    for (ω, ∂ω) in pairs(∂F)
        @test ∂ω ≈ getproperty(only(gs), ω)
    end

    samples = @inferred sample_from_inputs(layer, zeros(size(layer)..., 10^6))
    @test @inferred(mean_from_inputs(layer)) ≈ reshape(mean(samples; dims=3), size(layer)) rtol=0.1 atol=0.01
    @test @inferred(var_from_inputs(layer)) ≈ reshape(var(samples; dims=ndims(samples)), size(layer)) rtol=0.1
    @test @inferred(mean_abs_from_inputs(layer)) ≈ reshape(mean(abs.(samples); dims=ndims(samples)), size(layer)) rtol=0.1

    ∂F = @inferred RBMs.∂free_energy(layer)
    ∂E = @inferred RBMs.∂energy(layer, samples)
    @test length(∂F) == length(∂E)
    @test propertynames(∂F) == propertynames(∂E)
    for (∂f, ∂e) in zip(∂F, ∂E)
        @test ∂f ≈ ∂e rtol=0.1
    end

    gs = Zygote.gradient(layer) do layer
        sum(energies(layer, samples)) / size(samples)[end]
    end
    for (ω, ∂ω) in pairs(∂E)
        @test ∂ω ≈ getproperty(only(gs), ω)
    end
end

@testset "discrete layers ($Layer)" for Layer in (Binary, Spin, RBMs.Potts)
    N = (3, 4, 5)
    B = 13
    layer = Layer(randn(N...))
    x = bitrand(N..., B)
    @test energies(layer, x) ≈ -layer.θ .* x
    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    @test RBMs.∂free_energy(layer).θ ≈ only(gs).θ ≈ -mean_from_inputs(layer)
end

@testset "Binary" begin
    @testset "binary_rand" begin
        for θ in -5:5, u in 0.0:0.1:1.0
            @test (@. u * (1 + exp(-θ)) < 1) == @inferred RBMs.binary_rand(θ, u)
        end
        θ = randn(1000)
        u = rand(1000)
        @test RBMs.binary_rand.(θ, u) == (@. u * (1 + exp(-θ)) < 1)
    end

    layer = Binary(randn(7, 4, 5))
    var_from_inputs(layer) ≈ @. logistic(layer.θ) * logistic(-layer.θ)
    @test free_energies(layer) ≈ -log.(sum(exp.(layer.θ .* h) for h in 0:1))
    @test sort(unique(sample_from_inputs(layer))) == [0, 1]

    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θ ≈ only(gs).θ ≈ -mean_from_inputs(layer)
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)
end

@testset "Spin" begin
    layer = Spin(randn(7, 4, 5))
    @test free_energies(layer) ≈ -log.(sum(exp.(layer.θ .* h) for h in (-1, 1)))
    @test sort(unique(sample_from_inputs(layer))) == [-1, 1]

    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θ ≈ only(gs).θ ≈ -mean_from_inputs(layer)
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)
end

@testset "Potts" begin
    q = 3
    N = (4, 5)
    layer = RBMs.Potts(randn(q, N...))
    @test free_energies(layer) ≈ -log.(sum(exp.(layer.θ[h:h,:,:,:]) for h in 1:q))
    @test all(sum(mean_from_inputs(layer); dims=1) .≈ 1)
    # samples are proper one-hot
    @test sort(unique(sample_from_inputs(layer))) == [0, 1]
    @test all(sum(sample_from_inputs(layer); dims=1) .== 1)

    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θ ≈ only(gs).θ ≈ -mean_from_inputs(layer)
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)
end

@testset "Gaussian" begin
    N = (3, 4, 5)
    B = 7

    # bound γ away from zero to avoid numerical issues with quadgk, but allow negative
    layer = Gaussian(randn(N...), (rand(N...) .+ 1) .* (rand(Bool, N...) .- 0.5))

    x = randn(size(layer)..., B)
    @test energies(layer, x) ≈ @. abs(layer.γ) * x^2 / 2 - layer.θ * x

    function quad_free(θ::Real, γ::Real)
        Z, ϵ = quadgk(h -> exp(-RBMs.gauss_energy(θ, γ, h)), -Inf,  Inf)
        return -log(Z)
    end

    @test free_energies(layer) ≈ quad_free.(layer.θ, layer.γ) rtol=1e-6

    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    μ = mean_from_inputs(layer)
    ν = var_from_inputs(layer)
    μ2 = @. ν + μ^2
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θ ≈ only(gs).θ ≈ -μ
    @test ∂.γ ≈ only(gs).γ ≈ sign.(layer.γ) .* μ2/2
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)
end

@testset "ReLU" begin
    N = (3, 4, 5)
    B = 7

    # bound γ away from zero to avoid numerical issues with quadgk, but allow negative
    layer = ReLU(randn(N...), (rand(N...) .+ 1) .* (rand(Bool, N...) .- 0.5))

    x = abs.(randn(size(layer)..., B))
    @test energies(layer, x) ≈ energies(Gaussian(layer.θ, layer.γ), x)

    function quad_free(θ::Real, γ::Real)
        Z, ϵ = quadgk(h -> exp(-RBMs.relu_energy(θ, γ, h)), 0,  Inf)
        return -log(Z)
    end
    @test free_energies(layer) ≈ @. quad_free(layer.θ, layer.γ)

    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    μ = mean_from_inputs(layer)
    ν = var_from_inputs(layer)
    μ2 = @. ν + μ^2
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θ ≈ only(gs).θ ≈ -μ
    @test ∂.γ ≈ only(gs).γ ≈ sign.(layer.γ) .* μ2/2
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)
end

@testset "pReLU / xReLU / dReLU convert" begin
    N = (10, 7)
    B = 13
    x = randn(N..., B)

    drelu = dReLU(randn(N...), randn(N...), randn(N...), randn(N...))
    prelu = @inferred pReLU(drelu)
    xrelu = @inferred xReLU(drelu)
    @test drelu.θp ≈ dReLU(prelu).θp ≈ dReLU(xrelu).θp
    @test drelu.θn ≈ dReLU(prelu).θn ≈ dReLU(xrelu).θn
    @test abs.(drelu.γp) ≈ abs.(dReLU(prelu).γp)
    @test abs.(drelu.γn) ≈ abs.(dReLU(prelu).γn)
    @test abs.(drelu.γp) ≈ abs.(dReLU(xrelu).γp)
    @test abs.(drelu.γn) ≈ abs.(dReLU(xrelu).γn)
    @test energies(drelu, x) ≈ energies(prelu, x) ≈ energies(xrelu, x)
    @test free_energies(drelu) ≈ free_energies(prelu) ≈ free_energies(xrelu)
    @test mode_from_inputs(drelu) ≈ mode_from_inputs(prelu) ≈ mode_from_inputs(xrelu)
    @test mean_from_inputs(drelu) ≈ mean_from_inputs(prelu) ≈ mean_from_inputs(xrelu)
    @test mean_abs_from_inputs(drelu) ≈ mean_abs_from_inputs(prelu) ≈ mean_abs_from_inputs(xrelu)
    @test var_from_inputs(drelu) ≈ var_from_inputs(prelu) ≈ var_from_inputs(xrelu)

    prelu = pReLU(randn(N...), randn(N...), randn(N...), 2rand(N...) .- 1)
    drelu = @inferred dReLU(prelu)
    xrelu = @inferred xReLU(prelu)
    @test prelu.θ ≈ pReLU(drelu).θ ≈ pReLU(xrelu).θ
    @test abs.(prelu.γ) ≈ abs.(pReLU(drelu).γ) ≈ abs.(pReLU(xrelu).γ)
    @test prelu.Δ ≈ pReLU(drelu).Δ ≈ pReLU(xrelu).Δ
    @test prelu.η ≈ pReLU(drelu).η ≈ pReLU(xrelu).η
    @test energies(drelu, x) ≈ energies(prelu, x) ≈ energies(xrelu, x)
    @test free_energies(drelu) ≈ free_energies(prelu) ≈ free_energies(xrelu)
    @test mode_from_inputs(drelu) ≈ mode_from_inputs(prelu) ≈ mode_from_inputs(xrelu)
    @test mean_from_inputs(drelu) ≈ mean_from_inputs(prelu) ≈ mean_from_inputs(xrelu)
    @test mean_abs_from_inputs(drelu) ≈ mean_abs_from_inputs(prelu) ≈ mean_abs_from_inputs(xrelu)
    @test var_from_inputs(drelu) ≈ var_from_inputs(prelu) ≈ var_from_inputs(xrelu)

    xrelu = xReLU(randn(N...), randn(N...), randn(N...), randn(N...))
    drelu = @inferred dReLU(xrelu)
    prelu = @inferred pReLU(xrelu)
    @test xrelu.θ ≈ xReLU(drelu).θ ≈ xReLU(prelu).θ
    @test xrelu.Δ ≈ xReLU(drelu).Δ ≈ xReLU(prelu).Δ
    @test xrelu.ξ ≈ xReLU(drelu).ξ ≈ xReLU(prelu).ξ
    @test abs.(xrelu.γ) ≈ abs.(xReLU(prelu).γ)
    @test energies(drelu, x) ≈ energies(prelu, x) ≈ energies(xrelu, x)
    @test free_energies(drelu) ≈ free_energies(prelu) ≈ free_energies(xrelu)
    @test mode_from_inputs(drelu) ≈ mode_from_inputs(prelu) ≈ mode_from_inputs(xrelu)
    @test mean_from_inputs(drelu) ≈ mean_from_inputs(prelu) ≈ mean_from_inputs(xrelu)
    @test mean_abs_from_inputs(drelu) ≈ mean_abs_from_inputs(prelu) ≈ mean_abs_from_inputs(xrelu)
    @test var_from_inputs(drelu) ≈ var_from_inputs(prelu) ≈ var_from_inputs(xrelu)

    gauss = Gaussian(randn(N...), randn(N...))
    drelu = @inferred dReLU(gauss)
    prelu = @inferred pReLU(gauss)
    xrelu = @inferred xReLU(gauss)
    @test (
        energies(gauss, x) ≈ energies(drelu, x) ≈
        energies(prelu, x) ≈ energies(xrelu, x)
    )
    @test (
        free_energies(gauss) ≈ free_energies(drelu) ≈
        free_energies(prelu) ≈ free_energies(xrelu)
    )
    @test (
        mode_from_inputs(gauss) ≈ mode_from_inputs(drelu) ≈
        mode_from_inputs(prelu) ≈ mode_from_inputs(xrelu)
    )
    @test (
        mean_from_inputs(gauss) ≈ mean_from_inputs(drelu) ≈
        mean_from_inputs(prelu) ≈ mean_from_inputs(xrelu)
    )
    @test (
        mean_abs_from_inputs(gauss) ≈ mean_abs_from_inputs(drelu) ≈
        mean_abs_from_inputs(prelu) ≈ mean_abs_from_inputs(xrelu)
    )
    @test (
        var_from_inputs(gauss) ≈ var_from_inputs(drelu) ≈
        var_from_inputs(prelu) ≈ var_from_inputs(xrelu)
    )

    drelu = dReLU(randn(1), [0.0], randn(1), [Inf])
    relu = ReLU(drelu.θp, drelu.γp)
    x = rand(1, 100)
    @test energies(relu, x) ≈ energies(drelu, x)
    @test free_energies(relu) ≈ free_energies(drelu)
    @test mode_from_inputs(relu) ≈ mode_from_inputs(drelu)
    #@test mean_from_inputs(relu) ≈ mean_from_inputs(drelu)
    #@test mean_abs_from_inputs(relu)  ≈ mean_abs_from_inputs(drelu)
    #@test var_from_inputs(relu) ≈ var_from_inputs(drelu)
end

@testset "dReLU" begin
    N = (3, 5)
    B = 13
    x = randn(N..., B)

    # bound γ away from zero to avoid issues with quadgk
    layer = dReLU(
        randn(N...), randn(N...),
        (rand(N...) .+ 1) .* (rand(Bool, N...) .- 0.5),
        (rand(N...) .+ 1) .* (rand(Bool, N...) .- 0.5)
    )

    Ep = energy(ReLU( layer.θp, layer.γp), max.( x, 0))
    En = energy(ReLU(-layer.θn, layer.γn), max.(-x, 0))
    @test energy(layer, x) ≈ Ep + En
    @test iszero(energy(layer, zero(x)))

    function quad_free(θp::Real, θn::Real, γp::Real, γn::Real)
        Z, ϵ = quadgk(h -> exp(-RBMs.drelu_energy(θp, θn, γp, γn, h)), -Inf, Inf)
        return -log(Z)
    end
    @test free_energies(layer) ≈ quad_free.(layer.θp, layer.θn, layer.γp, layer.γn)

    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θp ≈ only(gs).θp
    @test ∂.θn ≈ only(gs).θn
    @test ∂.γp ≈ only(gs).γp
    @test ∂.γn ≈ only(gs).γn
    μ = mean_from_inputs(layer)
    ν = var_from_inputs(layer)
    μ2 = @. ν + μ^2
    @test ∂.θp + ∂.θn ≈ -μ
    @test ∂.γp .* sign.(layer.γp) + ∂.γn .* sign.(layer.γn) ≈ μ2/2
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)

    # check law of total variance
    inputs = randn(size(layer)..., 1000)
    ∂ = RBMs.∂free_energy(layer, inputs)
    h_ave = mean_from_inputs(layer, inputs)
    h_var = var_from_inputs(layer, inputs)
    μ = batchmean(layer, h_ave)
    ν_int = batchmean(layer, h_var)
    ν_ext = batchvar(layer, h_ave; mean = μ)
    ν = ν_int + ν_ext # law of total variance
    @test RBMs.grad2ave(layer, ∂) ≈ μ
    @test RBMs.grad2var(layer, ∂) ≈ ν
    μ1, ν1 = RBMs.total_meanvar_from_inputs(layer, inputs)
    @test μ1 ≈ μ ≈ RBMs.total_mean_from_inputs(layer, inputs)
    @test ν1 ≈ ν ≈ RBMs.total_var_from_inputs(layer, inputs)
end

@testset "pReLU" begin
    N = (3, 5, 7)
    layer = pReLU(randn(N...), randn(N...), randn(N...), 2rand(N...) .- 1)
    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θ ≈ only(gs).θ ≈ -mean_from_inputs(layer)
    @test ∂.γ ≈ only(gs).γ
    @test ∂.Δ ≈ only(gs).Δ
    @test ∂.η ≈ only(gs).η
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)
end

@testset "xReLU" begin
    N = (3, 5, 7)
    layer = xReLU(randn(N...), randn(N...), randn(N...), randn(N...))
    gs = Zygote.gradient(layer) do layer
        sum(free_energies(layer))
    end
    ∂ = RBMs.∂free_energy(layer)
    @test ∂.θ ≈ only(gs).θ ≈ -mean_from_inputs(layer)
    @test ∂.γ ≈ only(gs).γ
    @test ∂.Δ ≈ only(gs).Δ
    @test ∂.ξ ≈ only(gs).ξ
    @test RBMs.grad2ave(layer, ∂) ≈ mean_from_inputs(layer)
    @test RBMs.grad2var(layer, ∂) ≈ var_from_inputs(layer)
end
